#!/usr/bin/perl -w
use strict;
use Getopt::Long;
use Bio::Perl; # for reverse_complement_as_string
my $limit = 5;
my $flank = 2;
my $usage = <<END
cutFastq.pl -parsed hits -part seq2 -parsed file < fastq > cut.fastq

Given the output of parseAln.pl and the name of the part that lies
just after the barcode, cut out the barcode plus a few nucleotides of
flanking sequence (default $flank) to make a far smaller fastq file.

Ideally the hit to the part would extend to its beginning (position
1), but some slop is allowed (default $limit).

Hits to the part may be on either strand. If on the - strand, the SAM
file is reverse complemented. Only the first hit for that sequence is
considered.

Options:
-flank $flank -- the number of flanking nucleotides to extract
-limit $limit -- the limit on the amount of slop (default $limit)
-nosub -- remove the trailing /begin_end names from subreads
	(for PacBio fastq files; each template may have multiple
        entries with the same read name in the output)
-format fasta -- output fasta instead of fastq (default)
END
    ;

{
    my ($name, $parsefile);
    my $nosub;
    my $format = "fastq";
    die $usage
        unless GetOptions('part=s' => \$name,
                          'parsed=s' => \$parsefile,
                          'flank=i' => \$flank,
                          'limit=i' => \$limit,
                          'nosub' => \$nosub,
                          'format=s' => \$format )
        && defined $name
        && defined $parsefile;
    die "No such file: $parsefile\n" unless -e $parsefile;
    die "Flank must be nonnegative\n" unless $flank >= 0;
    die "Limit must be positive\n" unless $limit > 0;
    die "Unknown format" unless $format eq "fastq" || $format eq "fasta";

    my %readToPos = (); # read => the estimated beginning of the region just after the barcode, strand
    open(PARSE, "<", $parsefile) || die "Cannot read $parsefile";
    while(<PARSE>) {
        chomp;
        my ($read,$readLen,$readBeg,$readEnd,$strand,$ref,$refBegin,$refEnd,$mapqual) = split /\t/, $_, -1;
        die "Not enough fields in input file $parsefile" unless defined $mapqual;
        next unless $ref eq $name && $refBegin <= $limit;
        die "Invalid line strand $strand" unless $strand eq "+" || $strand eq "-";
        next if exists $readToPos{$read};
        # if $refBegin=1 then it's just readBeg or readEnd
        my $pos = $strand eq "+" ? $readBeg - $refBegin + 1 : $readEnd + $refBegin - 1;
        $readToPos{$read} = [ $pos, $strand ];
    }
    close(PARSE) || die "Error reading $parsefile";
    if (scalar(keys %readToPos) == 0) {
        print STDERR "Warning: No hits for $name in $parsefile\n";
        exit(0);
    }

    print STDERR "Truncating subread names to read names\n" if defined $nosub;

    my $nSkipShort = 0;
    my $nFound = 0;
    # process sequences
    while(my $header = <STDIN>) {
        chomp $header;
        $header =~ m/^@(.*)$/ || die "Input not in fastq format: no @";
        my $read = $1;
        my $seq = <STDIN>;
        chomp $seq;
        my $line3 = <STDIN>;
        chomp $line3;
        my $qual = <STDIN>;
        chomp $qual;
        die "Input not in fasta format: sequence contains non ACGTN:\n$seq\n"
            unless $seq =~ m/^[ACGTN]+$/;
        die "Quality score does not match length of sequence"
            unless length($qual) == length($seq);
        next unless exists $readToPos{$read};
        my ($pos,$strand) = @{ $readToPos{$read} };
        my $sign = $strand eq "+" ? 1 : -1;
        # If strand=+ and flank=0, we want pos-20 to pos-1
        my $end = $pos + ($flank-1) * $sign;
        my $begin = $pos + (-$flank-20) * $sign;
        ($begin,$end) = ($end,$begin) if $strand eq "-";
        $begin = 1 if $begin < 1;
        $end = length($seq) if $end > length($seq);
        if ($end-$begin+1 < 20) {
            $nSkipShort++;
            next;
        }
        my $subseq = substr($seq, $begin-1, $end-$begin+1);
        my $subqual = substr($qual, $begin-1, $end-$begin+1);
        if ($strand eq "-") {
            $subseq = reverse_complement_as_string($subseq);
            $subqual = reverse($subqual);
        }
        if (defined $nosub) {
            $read =~ s!/\d+_\d+$!!;
        }
        $nFound++;
        if ($format eq "fasta") {
            print ">$read\n$subseq\n";
        } else {
            print "$header\n$subseq\n$line3\n$subqual\n";
        }
    }
    print STDERR "Printed $nFound regions\n";
    print STDERR "Skipped " . $nSkipShort . " reads with too few nt before $name\n";
}
