#!/usr/bin/perl -w
use strict;
use Getopt::Long;

my $minMapQual = 10;
my $minBcScore = 30;
my $bcstrand = "-";

my $usage = <<END
identifyParts.pl -nparts 5 -parsed parts.parsed -blastn barcodes.blastn > table

The parts.parsed file should be from parseAln.pl and the references
should have names like *seqN, where N is an integer, or *partN.M, where *
is any string (or empty) and N and M are integers.

Builds a table of barcode, blastn score, parts 1:N, and read. A part
may be "mult" if it was ambiguous (i.e., well mapping reads to more
than one part) and will be empty if there was no alignment.

Options:
-mapqual -- minimum mapping quality (default $minMapQual)
-bcscore -- minimum blastn score for barcode alignment (default $minBcScore)
-bcstrand -- which strand the hits to the barcode should be on (default $bcstrand)
END
    ;

{
    my ($nparts, $parsedfile, $blastnfile);
    die $usage
        unless GetOptions('nparts=i' => \$nparts,
                          'parsed=s' => \$parsedfile,
                          'blastn=s' => \$blastnfile,
                          'mapqual=i' => \$minMapQual,
                          'bcscore=i' => \$minBcScore,
                          'bcstrand=s' => \$bcstrand )
        && @ARGV == 0
        && defined $nparts
        && defined $parsedfile
        && defined $blastnfile;
    die "Must have at least 1 part\n" unless $nparts >= 1;
    die "No such file: $parsedfile\n" unless -e $parsedfile;
    die "No such file: $blastnfile\n" unless -e $blastnfile;
    die "-bcstrand must be + or -" unless $bcstrand eq "+" || $bcstrand eq "-";

    my %readBc = (); # read to barcode, or empty for ambiguous
    my %readBcScore = (); # read to barcode's blastn score

    open(BLASTN, "<", $blastnfile) || die "Cannot read $blastnfile";
    my $nWrongStrand = 0;
    while(<BLASTN>) {
        chomp;
        my ($read,$barcode,$identity,$match,$mm,$gap,undef,undef,$bcBeg,$bcEnd,$evalue,$bits)
            = split /\t/, $_, -1;
        die "Not a valid line:\n$_\nfrom blastn"
            unless defined $bits
            && $bits =~ m/^ *[0-9.e]+$/
            && $bcBeg =~ m/^\d+$/
            && $bcEnd =~ m/^\d+$/
            && $read ne ""
            && $barcode ne "";
        next unless $bits >= $minBcScore;
        my $hitstrand = $bcBeg < $bcEnd ? "+" : "-";
        if ($hitstrand ne $bcstrand) {
            $nWrongStrand++;
            next;
        }
        # potentially ambiguous
        if (exists $readBcScore{$read} && $bits >= $readBcScore{$read} - 1) {
            if ($bits >= $readBcScore{$read} + 1) {
                # better hit, fall through
                ;
            } else {
                $readBc{$read} = "";
            }
        }
        if (!exists $readBc{$read} || $readBc{$read} ne "") {
            $readBc{$read} = $barcode;
            $readBcScore{$read} = $bits;
        }
    }
    close(BLASTN) || die "Error reading $blastnfile";

    my %bcReads = (); # barcode to list of reads
    while (my ($read,$bc) = each %readBc) {
        push @{ $bcReads{$bc} }, $read if $bc ne "";
    }

    my $nAmbiguous = scalar(grep { $_ eq ""} values(%readBc));
    print STDERR "Found " . scalar(keys %bcReads) . " quality barcodes for " . (scalar(keys %readBc) - $nAmbiguous) . " reads\n";
    print STDERR "\tambiguous barcodes for $nAmbiguous reads\n";
    print STDERR "\tignored $nWrongStrand quality hits with strand not '$bcstrand'\n";

    my %readPart = (); # read => part number => list of matching parts
    open(PARSED, "<", $parsedfile) || die "Error reading $parsedfile";
    while(<PARSED>) {
        chomp;
        my ($read,$readLen,$readBeg,$readEnd,$strand,$ref,$refB,$refE,$mapqual) = split /\t/, $_, -1;
        die "Not enough fields in \n$_\nfrom $parsedfile" unless defined $mapqual;
        next if $readLen eq "readLen" || $readLen eq "";
        die "Invalid map quality: $mapqual" unless $mapqual =~ m/^ *-?[0-9.]+$/;
        next unless $mapqual >= $minMapQual;
        next if $ref =~ m/seq[0-9]+$/;
        die "Invalid reference $ref" unless $ref =~ m/part([0-9]+)[.][0-9]+$/;
        my $ipart = $1;
        die "ref = part0.* not allowed" if $ipart <= 0;
        die "Too many parts: $ipart" unless $ipart <= $nparts;
        push @{ $readPart{$read}{$ipart} }, $ref;
    }
    close(PARSED) || die "Error reading $parsedfile";
    print STDERR "Read at least one part for " . scalar(keys %readPart) . " reads\n";

    my @partsheader = map { "parts".$_ } (1..$nparts);
    print join("\t", "barcode", "bcscore", @partsheader, "read")."\n";
    my $nBcOut = 0;
    foreach my $bc (sort keys %bcReads) {
        my @reads = sort @{ $bcReads{$bc} };
        my $useful = 0;
        foreach my $read (@reads) {
            next unless exists $readPart{$read};
            $useful = 1;
            my $hash = $readPart{$read};
            my @partstrings = ();
            foreach my $ipart (1..$nparts) {
                my $string = "";
                if (exists $hash->{$ipart}) {
                    my %refs = map { $_ => 1 } @{ $hash->{$ipart} };
                    if (scalar(keys %refs) > 1) {
                        $string = "mult";
                    } else {
                        $string = (keys %refs)[0];
                    }
                }
                push @partstrings, $string;
            }
            print join("\t", $bc, $readBcScore{$read}, @partstrings, $read)."\n";
        }
        $nBcOut++ if $useful;
    }
    print STDERR "Identified part(s) for $nBcOut different barcodes\n";
}
   
