#!/usr/bin/perl -w
# Given SAM format output of BWA, make a table of the form
#	read	readLen	readBeg	readEnd	strand	ref	refBegin	refEnd	mapqual
# with rBegin < rEnd, qBegin < qEnd, all 1-based, etc., sorted by rBegin
use strict;

# BWA alignment line => list of fields to output; or, a list containing the read only if a non-alignment
sub ParseLine($);
sub SummarizeRead($); # take list of list of fields and write to stdout

# S -- soft clipping (region of read that is shown in the $seq field but does not align)
# H -- hard clipping (region of read that does not align and is suppressed from $seq field)
# M/X -- match (or mismatch; use of X is optional)
# I -- insertion into reference
# D -- deletion from reference
# codes "=", "N", and "P" are not supported
my @CIGARCodes = split //, "SHMXID";
my %CIGARCodes = map {$_ => 1} @CIGARCodes;

{
    die "Run parseAln.pl as a filter\n" unless @ARGV ==0;
    print join("\t", qw{read readLen readBeg readEnd strand ref refBegin refEnd mapqual})."\n";
    my $currentread;
    my @alns = ();
    while(my $line = <STDIN>) {
        next if $line =~ m/^@/;
        my $aln = ParseLine($line);
        my $read = $aln->[0];
        if (defined $currentread && $read ne $currentread) {
            SummarizeRead(\@alns);
            @alns = ();
        }
        $currentread = $read;
        push @alns, $aln;
    }
    SummarizeRead(\@alns) if scalar(@alns) > 0;
}

sub ParseLine($) {
    my ($line) = @_;
    chomp $line;
    my ($read,$flag,$refname,$refBegin,$mapq,$cigar,$rnext,$pnext,$tlen,$seq,$qual) = split /\t/, $line, -1;
    return [ $read ] if ($flag & 4); # unmapped
    my $strand = ($flag & 16) ? "-" : "+";
    my $pairedend = ($flag & 1);
    die "Cannot handle paired-end alignment for $read" if $pairedend;
    die "Invalid pos field for $read $refname" if $refBegin eq "*" || $refBegin == 0;
    # supplementary is given by $flag & 2048; not used here

    # parse cigar
    my @cigar = (); # list of code,nchars
    my $cigarleft = $cigar;
    while ($cigarleft =~ m/^(\d+)([A-Z=])(.*)$/) {
        my ($len, $code) = ($1,$2);
        $cigarleft = $3;
        push @cigar, [$code,$len];
    }
    die "Cannot parse remaining CIGAR string $cigarleft for $read $refname" unless $cigarleft eq "";
    die "Empty cigar string is not allowed for $read $refname" if @cigar == 0;
    my $ShownReadLen = 0;
    my $ActualReadLen = 0;
    my $MatchReadLen = 0;
    my $MatchRefLen = 0;
    my $alnyet = 0; # set to 1 once no longer clipping
    my $leftclip = 0;
    foreach my $row (@cigar) {
        my ($code,$len) = @$row;
        die "Cannot not handle code CIGAR $code code for $read $refname"
            unless exists $CIGARCodes{$code};
        $MatchRefLen += $len if $code eq "M" || $code eq "X" || $code eq "D";
        $MatchReadLen += $len if $code eq "M" || $code eq "X" || $code eq "I";
        $ShownReadLen += $len if $code ne "D" && $code ne "H"; # assuming only SHMXID is allowed
        $ActualReadLen += $len if $code ne "D"; # assuming only SHMXID is allowed
        $leftclip += $len if ($code eq "S" || $code eq "H") && !$alnyet;
        $alnyet = 1 unless $code eq "S" || $code eq "H";
    }
    die "Non-matching length between CIGAR and shown sequence for $read $refname $strand $refBegin $cigar"
        if $ShownReadLen != length($seq);
             
    my $rBegin = $strand eq "+" ? 1+$leftclip : $ActualReadLen - $leftclip;
    die "Invalid rBegin $rBegin (leftclip $leftclip actual $ActualReadLen strand $strand) for $read $refname $cigar" unless $rBegin >= 1 && $rBegin <= $ActualReadLen;
    my $rEnd = $strand eq "+" ? $rBegin + $MatchReadLen-1 : $rBegin + 1 - $MatchReadLen;
    ($rBegin,$rEnd) = ($rEnd,$rBegin) if ($strand eq "-");
    die "Invalid rEnd $rEnd left $leftclip actual $ActualReadLen shown $ShownReadLen strand $strand for $read $refname $cigar" unless $rEnd >= 1 && $rEnd <= $ActualReadLen;
    return [ $read, $ActualReadLen, $rBegin, $rEnd, $strand,
             $refname, $refBegin, $refBegin + $MatchRefLen - 1,
             $mapq ];
}

sub SummarizeRead($) {
    my ($list) = @_;
    foreach my $row (@$list) {
        if (scalar(@$row) == 1) {
            print join("\t", $row->[0], "", "", "", "", "", "", "", "")."\n";
        } else {
            print join("\t", @$row)."\n";
        }
    }
}

