#!/usr/bin/perl -w
use strict;
use Getopt::Long;

my $coverage = 0.98;
my $minN = 2;
my $preseq = "CAGCGTACG";
my $postseq = "AGAGACCTC";
my $flank = 2;

my $usage = <<END
Usage: topCodesToFna.pl -codes p.codes > p.fna

Turns a codes file (from MultiCodes.pl) into a fasta file that
contains only the popular barcodes, along with some flanking
sequence. It chooses a threshold so that the barcodes that
are included cover almost all ($coverage) of the reads, but with a
minimum of $minN reads.

Not compatible with multiplexed runs of MultiCodes.pl

Options:
-cover $coverage -- coverage of reads to aim for
-min $minN -- minimum reads per barcode
	(min only matters if coverage at this threshold is too low)
-flank $flank -- number of flanking nt on each side to show
-preseq $preseq -- sequence upstream of barcode
-postseq $postseq -- sequence downstream of barcode
END
    ;

{
    my $codesfile;
    die $usage
        unless GetOptions('codes=s' => \$codesfile,
                           'cover=f' => \$coverage,
                           'min=i' => $minN,
                          'flank=i' => \$flank,
                          'preseq=s' => \$preseq,
                          'postseq=s' => \$postseq )
        && @ARGV == 0
        && defined $codesfile;
    die "Coverage must be between 0 and 1\n" unless $coverage > 0 && $coverage <= 1;
    die "-flank argument must be nonnegative\n" unless $flank >= 0;
    die "preseq is too short or flank is too large\n"
        if $flank > length($preseq);
    die "postseq is too short or flank is too large\n"
        if $flank > length($postseq);

    my @nToCodes = (); # nreads to number of codes
    my $nReads = 0;
    my $nCodes = 0;
    open(CODES, "<", $codesfile) || die "Cannot read $codesfile\n";
    while(<CODES>) {
        chomp;
        my @F = split /\t/, $_, -1;
        die "Too many columns in $codesfile -- this script does not work on multiplexed runs"
            unless scalar(@F) <= 2;
        my ($barcode,$N) = @F;
        next if $barcode eq "barcode";
        die "Invalid count $N" unless $N =~ m/^\d+$/ && $N >= 1;
        die "Invalid barcode $barcode" unless $barcode =~ m/^[ACGTN]+$/;
        # ignore ambiguous barcodes if they occur
        next if $barcode =~ m/N/;
        push @{ $nToCodes[$N] }, $barcode;
        $nReads += $N;
        $nCodes++;
    }
    close(CODES) || die "Error reading $codesfile";
    print STDERR "Read $nReads reads for $nCodes unambiguous barcodes\n";
    my $n1 = scalar(@{ $nToCodes[1] });
    my $n2 = scalar(@{ $nToCodes[2] });
    print STDERR "\tSingletons $n1 Two-reads $n2 barcodes\n";

    my $nReadsSoFar = 0;
    my $nThresh;
    my $nCodesShown = 0;
    for ($nThresh = scalar(@nToCodes)-1; $nThresh >= 1 && $nThresh >= $minN; $nThresh--) {
        my $list = $nToCodes[$nThresh];
        $list = [] if !defined $list;
        $nCodesShown += scalar(@$list);
        $nReadsSoFar += scalar(@$list) * $nThresh;
        last if $nReadsSoFar >= $nReads * $coverage;
    }
    print STDERR "Using threshold $nThresh covering " . sprintf("%.3f", $nReadsSoFar/$nReads) . " of reads\n";
    for (my $n = scalar(@nToCodes)-1; $n >= $nThresh; $n--) {
        my $list = $nToCodes[$n];
        $list = [] if !defined $list;
        foreach my $barcode (sort @$list) {
            my $seq = join("",
                           substr($preseq, length($preseq)-$flank, $flank),
                           $barcode,
                           substr($postseq, 0, $flank) );
            print ">$barcode\n$seq\n";
        }
    }
    print STDERR "Printed $nCodesShown barcodes\n";
}
